Object.defineProperty(exports, "__esModule", { value: true });
exports.Filter = void 0;
var util = require('../util');
var Model_1 = require("./Model");
var defaultFns = require("./defaultFns");
Model_1.Model.INITS.push(function (model) {
    model.root._filters = new Filters(model);
    model.on('all', filterListener);
    function filterListener(segments, event) {
        var passed = event.passed;
        var map = model.root._filters.fromMap;
        for (var path in map) {
            var filter = map[path];
            if (passed.$filter === filter)
                continue;
            if (util.mayImpact(filter.segments, segments) ||
                (filter.inputsSegments && util.mayImpactAny(filter.inputsSegments, segments))) {
                filter.update(passed);
            }
        }
    }
});
function parseFilterArguments(model, args) {
    var fn = args.pop();
    var options, inputPaths;
    var path = model.path(args.shift());
    var last = args[args.length - 1];
    if (!model.isPath(last) && !Array.isArray(last)) {
        options = args.pop();
    }
    if (args.length === 1 && Array.isArray(args[0])) {
        // inputPaths provided as one array:
        //   model.filter(inputPath, [inputPath1, inputPath2], fn)
        inputPaths = args[0];
    }
    else {
        // inputPaths provided as var-args:
        //   model.filter(inputPath, inputPath1, inputPath2, fn)
        inputPaths = args;
    }
    var i = inputPaths.length;
    while (i--) {
        inputPaths[i] = model.path(inputPaths[i]);
    }
    return {
        path: path,
        inputPaths: (inputPaths.length) ? inputPaths : null,
        options: options,
        fn: fn
    };
}
Model_1.Model.prototype.filter = function () {
    var args = Array.prototype.slice.call(arguments);
    var parsed = parseFilterArguments(this, args);
    return this.root._filters.add(parsed.path, parsed.fn, null, parsed.inputPaths, parsed.options);
};
Model_1.Model.prototype.sort = function () {
    var args = Array.prototype.slice.call(arguments);
    var parsed = parseFilterArguments(this, args);
    return this.root._filters.add(parsed.path, null, parsed.fn || 'asc', parsed.inputPaths, parsed.options);
};
Model_1.Model.prototype.removeAllFilters = function (subpath) {
    var segments = this._splitPath(subpath);
    this._removeAllFilters(segments);
};
Model_1.Model.prototype._removeAllFilters = function (segments) {
    var filters = this.root._filters.fromMap;
    for (var from in filters) {
        if (util.contains(segments, filters[from].fromSegments)) {
            filters[from].destroy();
        }
    }
};
var FromMap = /** @class */ (function () {
    function FromMap() {
    }
    return FromMap;
}());
var Filters = /** @class */ (function () {
    function Filters(model) {
        this.model = model;
        this.fromMap = new FromMap();
    }
    Filters.prototype.add = function (path, filterFn, sortFn, inputPaths, options) {
        return new Filter(this, path, filterFn, sortFn, inputPaths, options);
    };
    ;
    Filters.prototype.toJSON = function () {
        var out = [];
        for (var from in this.fromMap) {
            var filter = this.fromMap[from];
            // Don't try to bundle if functions were passed directly instead of by name
            if (!filter.bundle)
                continue;
            var args = [from, filter.path, filter.filterName, filter.sortName, filter.inputPaths];
            if (filter.options)
                args.push(filter.options);
            out.push(args);
        }
        return out;
    };
    ;
    return Filters;
}());
var Filter = /** @class */ (function () {
    function Filter(filters, path, filterFn, sortFn, inputPaths, options) {
        this.filters = filters;
        this.model = filters.model.pass({ $filter: this });
        this.path = path;
        this.segments = path.split('.');
        this.filterName = null;
        this.sortName = null;
        this.bundle = true;
        this.filterFn = null;
        this.sortFn = null;
        this.inputPaths = inputPaths;
        this.inputsSegments = null;
        if (inputPaths) {
            this.inputsSegments = [];
            for (var i = 0; i < this.inputPaths.length; i++) {
                var segments = this.inputPaths[i].split('.');
                this.inputsSegments.push(segments);
            }
        }
        this.options = options;
        this.skip = options && options.skip;
        this.limit = options && options.limit;
        if (filterFn)
            this.filter(filterFn);
        if (sortFn)
            this.sort(sortFn);
        this.idsSegments = null;
        this.from = null;
        this.fromSegments = null;
    }
    Filter.prototype.filter = function (fn) {
        if (typeof fn === 'function') {
            this.filterFn = fn;
            this.bundle = false;
            return this;
        }
        else if (typeof fn === 'string') {
            this.filterName = fn;
            this.filterFn = this.model.root._namedFns[fn] || defaultFns[fn];
            if (!this.filterFn) {
                throw new TypeError('Filter function not found: ' + fn);
            }
        }
        return this;
    };
    ;
    Filter.prototype.sort = function (fn) {
        if (!fn)
            fn = 'asc';
        if (typeof fn === 'function') {
            this.sortFn = fn;
            this.bundle = false;
            return this;
        }
        else if (typeof fn === 'string') {
            this.sortName = fn;
            this.sortFn = this.model.root._namedFns[fn] || defaultFns[fn];
            if (!this.sortFn) {
                throw new TypeError('Sort function not found: ' + fn);
            }
        }
        return this;
    };
    ;
    Filter.prototype._slice = function (results) {
        if (this.skip == null && this.limit == null)
            return results;
        var begin = this.skip || 0;
        // A limit of zero is equivalent to setting no limit
        var end;
        if (this.limit)
            end = begin + this.limit;
        return results.slice(begin, end);
    };
    ;
    Filter.prototype.getInputs = function () {
        if (!this.inputsSegments)
            return;
        var inputs = [];
        for (var i = 0, len = this.inputsSegments.length; i < len; i++) {
            var input = this.model._get(this.inputsSegments[i]);
            inputs.push(input);
        }
        return inputs;
    };
    ;
    Filter.prototype.callFilter = function (items, key, inputs) {
        var item = items[key];
        return (inputs) ?
            this.filterFn.apply(this.model, [item, key, items].concat(inputs)) :
            this.filterFn.call(this.model, item, key, items);
    };
    ;
    Filter.prototype.ids = function () {
        var items = this.model._get(this.segments);
        var ids = [];
        if (!items)
            return ids;
        if (Array.isArray(items)) {
            throw new Error('model.filter is not currently supported on arrays');
        }
        if (this.filterFn) {
            var inputs = this.getInputs();
            for (var key in items) {
                if (items.hasOwnProperty(key) && this.callFilter(items, key, inputs)) {
                    ids.push(key);
                }
            }
        }
        else {
            ids = Object.keys(items);
        }
        var sortFn = this.sortFn;
        if (sortFn) {
            ids.sort(function (a, b) {
                return sortFn(items[a], items[b]);
            });
        }
        return this._slice(ids);
    };
    ;
    Filter.prototype.get = function () {
        var items = this.model._get(this.segments);
        var results = [];
        if (Array.isArray(items)) {
            throw new Error('model.filter is not currently supported on arrays');
        }
        if (this.filterFn) {
            var inputs = this.getInputs();
            for (var key in items) {
                if (items.hasOwnProperty(key) && this.callFilter(items, key, inputs)) {
                    results.push(items[key]);
                }
            }
        }
        else {
            for (var key in items) {
                if (items.hasOwnProperty(key)) {
                    results.push(items[key]);
                }
            }
        }
        if (this.sortFn)
            results.sort(this.sortFn);
        return this._slice(results);
    };
    ;
    Filter.prototype.update = function (pass) {
        var ids = this.ids();
        this.model.pass(pass, true)._setArrayDiff(this.idsSegments, ids);
    };
    ;
    Filter.prototype.ref = function (from) {
        from = this.model.path(from);
        this.from = from;
        this.fromSegments = from.split('.');
        this.filters.fromMap[from] = this;
        this.idsSegments = ['$filters', from.replace(/\./g, '|')];
        this.update();
        return this.model.refList(from, this.path, this.idsSegments.join('.'));
    };
    ;
    Filter.prototype.destroy = function () {
        delete this.filters.fromMap[this.from];
        this.model._removeRef(this.idsSegments);
        this.model._del(this.idsSegments);
    };
    ;
    return Filter;
}());
exports.Filter = Filter;
