import { Model } from './Model';
import { CollectionCounter } from './CollectionCounter';
import { Query } from './Query';
/**
 * A path string, a `Model`, or a `Query`.
 */
export type Subscribable = string | Model<unknown> | Query<unknown>;
declare module './Model' {
    interface Model<T> {
        /**
         * Retrieve data from the server, loading it into the model.
         *
         * @param items
         * @param cb
         *
         * @see https://derbyjs.github.io/derby/models/backends#loading-data-into-a-model
         */
        fetch(items: Subscribable[], cb?: ErrorCallback): Model<T>;
        fetch(item: Subscribable, cb?: ErrorCallback): Model<T>;
        fetch(cb?: ErrorCallback): Model<T>;
        fetchPromised(items: Subscribable[]): Promise<void>;
        fetchPromised(item: Subscribable): Promise<void>;
        fetchPromised(): Promise<void>;
        fetchDoc(collecitonName: string, id: string, callback?: ErrorCallback): void;
        fetchDocPromised(collecitonName: string, id: string): Promise<void>;
        fetchOnly: boolean;
        /**
         * Retrieve data from the server, loading it into the model. In addition,
         * subscribe to the items, such that updates from any other client will
         * automatically get reflected in this client's model.
         *
         * Any item that's already subscribed will not result in a network call.
         *
         * @param items
         * @param cb
         *
         * @see https://derbyjs.github.io/derby/models/backends#loading-data-into-a-model
         */
        subscribe(items: Subscribable[], cb?: ErrorCallback): Model<T>;
        subscribe(item: Subscribable, cb?: ErrorCallback): Model<T>;
        subscribe(cb?: ErrorCallback): Model<T>;
        subscribePromised(items: Subscribable[]): Promise<void>;
        subscribePromised(item: Subscribable): Promise<void>;
        subscribePromised(): Promise<void>;
        subscribeDoc(collecitonName: string, id: string, callback?: ErrorCallback): void;
        subscribeDocPromised(collecitonName: string, id: string): Promise<void>;
        /**
         * The reverse of `#fetch`, marking the items as no longer needed in the
         * model.
         *
         * @param items
         * @param cb
         *
         * @see https://derbyjs.github.io/derby/models/backends#loading-data-into-a-model
         */
        unfetch(items: Subscribable[], cb?: ErrorCallback): Model<T>;
        unfetch(item: Subscribable, cb?: ErrorCallback): Model<T>;
        unfetch(cb?: ErrorCallback): Model<T>;
        unfetchPromised(items: Subscribable[]): Promise<void>;
        unfetchPromised(item: Subscribable): Promise<void>;
        unfetchPromised(): Promise<void>;
        unfetchDoc(collecitonName: string, id: string, callback?: (err?: Error, count?: number) => void): void;
        unfetchDocPromised(collecitonName: string, id: string): Promise<void>;
        unloadDelay: number;
        /**
         * The reverse of `#subscribe`, marking the items as no longer needed in the
         * model.
         *
         * @param items
         * @param cb
         *
         * @see https://derbyjs.github.io/derby/models/backends#loading-data-into-a-model
         */
        unsubscribe(items: Subscribable[], cb?: ErrorCallback): Model<T>;
        unsubscribe(item: Subscribable, cb?: ErrorCallback): Model<T>;
        unsubscribe(cb?: ErrorCallback): Model<T>;
        unsubscribePromised(): Promise<void>;
        unsubscribeDoc(collecitonName: string, id: string, callback?: (err?: Error, count?: number) => void): void;
        unsubscribeDocPromised(collecitonName: string, id: string): Promise<void>;
        _fetchedDocs: CollectionCounter;
        _forSubscribable(argumentsObject: any, method: any): void;
        _hasDocReferences(collecitonName: string, id: string): boolean;
        _maybeUnloadDoc(collecitonName: string, id: string): void;
        _subscribedDocs: CollectionCounter;
    }
}
